require_relative 'test_pb'

module Test
  def self.parse(query)
    input = File.open('proto.bin', 'rb')
    result = input.read
    stderr = []

    begin
      result = Test::ParseResult.decode(result)
    rescue Google::Protobuf::ParseError => e
      raise e
    end

    warnings = []

    Test::ParserResult.new(query, result, warnings)
  end

  class ParserResult
    attr_reader :query
    attr_reader :tree
    attr_reader :warnings

    def initialize(query, tree, warnings = [])
      @query = query
      @tree = tree
      @warnings = warnings
      @tables = nil
    end

    def tables
      tables_with_details.map { |t| t[:name] }.uniq
    end

    def tables_with_details
      load_objects! if @tables.nil?
      @tables
    end

    protected

    def load_objects! # rubocop:disable Metrics/CyclomaticComplexity
      @tables = [] # types: select, dml, ddl
      @cte_names = []
      @aliases = {}
      @functions = [] # types: call, ddl

      statements = @tree.stmts.dup.to_a.map(&:stmt)
      from_clause_items = [] # types: select, dml, ddl
      subselect_items = []

      loop do
        statement = statements.shift
        if statement
          case statement.node
          when :list
            statements += statement.list.items
          # The following statement types do not modify tables and are added to from_clause_items
          # (and subsequently @tables)
          when :select_stmt
            subselect_items.concat(statement.select_stmt.target_list)
            subselect_items << statement.select_stmt.where_clause if statement.select_stmt.where_clause
            subselect_items.concat(statement.select_stmt.sort_clause.collect { |h| h.sort_by.node })
            subselect_items.concat(statement.select_stmt.group_clause)
            subselect_items << statement.select_stmt.having_clause if statement.select_stmt.having_clause

            if statement.select_stmt.with_clause
              cte_statements, cte_names = statements_and_cte_names_for_with_clause(statement.select_stmt.with_clause)
              @cte_names.concat(cte_names)
              statements.concat(cte_statements)
            end
          end
        end

        next_item = subselect_items.shift
        if next_item
          case next_item.node
          when :a_expr
            %w[lexpr rexpr].each do |side|
              elem = next_item.a_expr.public_send(side)
              next unless elem
              if elem.is_a?(Array) # FIXME: this needs to traverse a list
                subselect_items += elem
              else
                subselect_items << elem
              end
            end
          when :bool_expr
            subselect_items.concat(next_item.bool_expr.args)
          when :res_target
            subselect_items << next_item.res_target.val
          end
        end

        break if subselect_items.empty? && statements.empty?
      end

      @tables.uniq!
    end
  end
end

sql = <<~EOF
SELECT 1 FROM "namespaces" WHERE 1 AND 2 AND 3
EOF

GC.stress = true

tables = Test::parse(sql).tables
